import torch
from torch import nn
from torch.nn import BCELoss, BCEWithLogitsLoss

class WDLLoss(nn.Module):
    """ Implements Win-Draw-Loss loss function"""
    def __init__(self, input_scale: float = 410.0, target_scale: float = 410.0):
        super().__init__()
        self.input_scale = input_scale
        self.target_scale = target_scale
        self.bce = BCEWithLogitsLoss()

    def forward(self, input, target):
        return self.bce(input / self.input_scale, torch.sigmoid(target / self.target_scale))




import numpy as np
from potatorch.optimization import BayesianOptimizer

class HyperOptimizer():
  def __init__(self, sweep_config):
    self.sweep_config = sweep_config
    self.minimize = sweep_config['metric']['goal'] == 'minimize'
    self.metric = sweep_config['metric']['name']
    self.method = sweep_config['method']
    self.config = sweep_config.get('fixed', {})
    self.best_config = self.config.copy()
    self.best_metric = np.inf if self.minimize else -np.inf

  def _assign_state(self, state):
    """ Set the current hyperparameters configuration """
    for k in state.keys():
      self.config[k] = state[k]

  def _random_state(self, rng, parameters, ranges):
    """ Generate a random configuration of hypeparameters following the
        distributions indicated into the sweep configuration.
    """
    state = {}
    for i in range(len(parameters)):
      p = parameters[i]
      r = ranges[i]
      dist = r.get('distribution') # Determine random distribution to use

      # Check if the distribution is finite (and provided as a list of values)
      if 'values' in r and self.method == 'grid':
        # Uniform sampling within r['values']
        v = rng.choice(r['values'], replace=True)
      elif 'min' in r and 'max' in r:
        # Uniform sampling between min and max
        low = r['min']
        high = r['max']
        # Discrete distribution
        if isinstance(low, int) or isinstance(high, int):
          v = rng.integers(low, high, endpoint=True, size=None)
        # Continuos distribution
        else:
          v = rng.random() # Sample from U(0, 1)
          v = (high - low) * v + low # Convert into U(low, high), low <= v < high
      else:
        raise Exception('Invalid distribution for {}'.format(p))

      # Log distribution:
      # must convert the sampled value from logarithmic form into normal form
      # TODO: set logarithm base as a configuration parameter
      if dist and dist.startswith('log_'):
        v = 10.0**v
      
      assert v is not None
      state[p] = v
    return state

  def _grid_search(self, score_function, parameters, ranges):
    """ Perform am exhaustive grid search over the hyperparameters' space """
    # Base case: call the score function with the current parameters
    if len(parameters) == 0:
      print('Evaluating configuration: {}'.format(self.config))
      target = score_function(self.config)[self.metric]
      if (self.minimize and target < self.best_metric) or (not self.minimize and target > self.best_metric):
        self.best_metric = target
        self.best_config = self.config.copy()
      return
  
    # Recursively test every possible configuration of parameters
    for v in ranges[0]['values']:
      self.config[parameters[0]] = v
      self._grid_search(score_function, 
                        parameters[1:],
                        ranges[1:])

  def _random_search(self, score_function, parameters, ranges, iterations=30):
    """ Perform a random search over the hyperparameters' space """
    if iterations is None or iterations <= 0: 
      raise Exception('Must provide a positive number of iterations')

    # TODO: provide probability distributions other than uniform
    # Random number generator
    # Note: this method always uses a different initial random state;
    #       if reproducibility is required, pass a seed to default_rnd(seed)
    rng = np.random.default_rng()

    for _ in range(iterations):
      # Generate a random configuration of hyperparameters
      state = self._random_state(rng, parameters, ranges)
      self._assign_state(state)
      # Evaluate random hyperparameters configuration
      print('Evaluating configuration: {}'.format(self.config))
      target = score_function(self.config)[self.metric]
      if (self.minimize and target < self.best_metric) or (not self.minimize and target > self.best_metric):
        self.best_metric = target
        self.best_config = self.config.copy()

  def _bayesian(self, score_function, parameters, ranges, iterations=30, sample_size=500):
    """ Perform a bayesian search over the hyperparameters' space """
    # Random number generator
    rng = np.random.default_rng()

    state = self._random_state(rng, parameters, ranges)
    self._assign_state(state)
    
    # TODO encanpsulate this print into a lambda with score_function
    print('Evaluating configuration: {}'.format(self.config))
    score = score_function(self.config)[self.metric]
    opt = BayesianOptimizer(state, score, self.minimize) 
    self.best_metric = score
    self.best_config = self.config.copy()
    self.optimizer = opt

    history = [score]
    best_iteration = 0
    for i in range(1, iterations):
      # Sample `sample_size` configurations of hyperparameters
      sample = np.array([self._random_state(rng, parameters, ranges) for _ in range(sample_size)])
      new_state = opt.optimize(sample) # Get the next hyperparameters to test
      self._assign_state(new_state) # Update bayesian model

      print('Evaluating configuration: {}'.format(self.config))
      score = score_function(self.config)[self.metric]
      opt.update(new_state, score)

      if (self.minimize and score < self.best_metric) or (not self.minimize and score > self.best_metric):
        best_iteration = i
        self.best_metric = score
        self.best_config = self.config.copy()

      history.append(score)

  def optimize(self, score_function, return_error=False, window_iterations=1):
    """ Perform hyperparameters optimization.
        The search method is indicated under the 'method' key of the
        sweep configuration dictionary,

        returns: the best configuration of hyperparameters found by the method
    """
    parameters = list(self.sweep_config['parameters'].keys())
    values = list(self.sweep_config['parameters'].values())
    iterations = self.sweep_config.get('iterations')
    assert iterations or (self.method == 'grid'), 'When performing bayesian or random optimization you must specify the number of iterations'

    default_values = values

    for i in range(window_iterations):
        if self.method == 'grid':
          self._grid_search(score_function, parameters, values)
        elif self.method == 'bayes':
          self._bayesian(score_function, parameters, values, iterations, sample_size=100)
        elif self.method == 'random':
          self._random_search(score_function, parameters, values, iterations)
        else:
          raise Exception('No search method provided')

        if window_iterations > 1:
          w = 0.9**(i+1)
          centers = self.best_config
          values = self._narrow_ranges(values, default_values, parameters, centers, window_size=w)
          print(f'Best configuration so far: {centers}')
          print(f'New ranges: {values}')

    if return_error:
        return self.best_config, self.best_metric
    return self.best_config

  def _narrow_ranges(self, ranges, default, parameters, centers, window_size=0.2):
    new_ranges = ranges.copy()
    for i in range(len(parameters)):
      p = parameters[i]
      c = centers[p]

      low = new_ranges[i]['min']
      high = new_ranges[i]['max']
      new_min = max(c * (1 - window_size), default[i]['min'])
      new_max = min(c * (1 + window_size), default[i]['max'])
      if isinstance(low, int) or isinstance(high, int):
          new_min = int(new_min)
          new_max = int(new_max)

      new_ranges[i]['min'] = new_min
      new_ranges[i]['max'] = new_max

    return new_ranges



import torch
from torch.utils.data import Dataset

class FilteredDataset(Dataset):
    """ Wrapper class that filters a dataset with a provided filtering function.

        - `filter_fn` is such that:
           new_dataset = [d for d in old_dataset if filter_fn(d)]
    """
    def __init__(self, dataset, filter_fn):
        self.dataset = dataset
        self.indices = [idx for idx in range(len(dataset)) if filter_fn(dataset[idx])]

    def __len__(self):
        return len(self.indices)

    def __getitem__(self, idx):
        return self.dataset[self.indices[idx]]



import torch
from torch import nn
from torch.jit import Final
from torchvision.ops import SqueezeExcitation

class DepthwiseSeparable2d(nn.Module):
    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size=(2,2),
                 stride=1,
                 dilation=1,
                 activation=nn.ReLU(),
                 padding='same',
                 normalize=True,
                 pool=False,
                 se_layer=False):
        super(DepthwiseSeparable2d, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.activation = activation
        self.padding = padding
        self.normalize = normalize
        self.pool = pool

        # depthwise
        self.depthwise = nn.Conv2d(
                self.in_channels, self.in_channels,
                kernel_size=self.kernel_size,
                stride=stride,
                dilation=dilation,
                bias=(not self.normalize),
                padding=self.padding,
                groups=self.in_channels
                )
        # pointwise
        self.pointwise = nn.Conv2d(
                self.in_channels, self.out_channels,
                kernel_size=1,
                stride=stride,
                dilation=dilation,
                bias=(not self.normalize)
                )

        if self.normalize:
            self.norm_layer = nn.BatchNorm2d(self.out_channels)
            # self.norm_layer = nn.GroupNorm(1, self.out_channels)
        if self.pool:
            self.pool_layer = nn.MaxPool2d(kernel_size=(2,2))

    def forward(self, x):
        x = self.depthwise(x)

        if self.activation:
            x = self.activation(x)

        x = self.pointwise(x)

        if self.normalize:
            x = self.norm_layer(x)

        # TODO: try activation also after pointwise conv

        if self.pool:
            x = self.pool_layer(x)
        return x


class Conv2dBlock(nn.Module):
    normalize: Final[bool]
    pool: Final[bool]
    se_layer: Final[nn.Module]

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 kernel_size=(2,2),
                 stride=1,
                 dilation=1,
                 activation=nn.ReLU(),
                 padding='same',
                 normalize=True,
                 pool=False,
                 se_layer=False):
        super(Conv2dBlock, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.activation = activation
        self.padding = padding
        self.normalize = normalize
        self.pool = pool
        self.se_layer = se_layer

        self.conv = nn.Conv2d(
                self.in_channels, self.out_channels,
                kernel_size=self.kernel_size,
                stride=stride,
                dilation=dilation,
                bias=(not self.normalize),
                padding=self.padding
                )
        if self.normalize:
            self.norm_layer = nn.BatchNorm2d(self.out_channels)
            # self.norm_layer = nn.GroupNorm(1, self.out_channels)

        if self.se_layer:
            squeeze_channels = max(1, self.out_channels // 12)
            self.se = SqueezeExcitation(self.out_channels, squeeze_channels)

        if self.pool:
            self.pool_layer = nn.MaxPool2d(kernel_size=(2,2))


    def forward(self, x):
        x = self.conv(x)
        if self.activation is not None:
            x = self.activation(x)
        if self.normalize:
            x = self.norm_layer(x)
        if self.se_layer:
            x = self.se(x)
        if self.pool:
            x = self.pool_layer(x)
        return x



import wandb
import torch
import os

def download_wandb_checkpoint(run_path, filename, device='cuda'):
    api = wandb.Api()

    run = api.run(run_path)
    run.file(filename).download(replace=True)
    checkpoint = torch.load(filename, map_location=torch.device(device))
    return checkpoint

def save_wandb_file(path):
    wandb.save(path, base_path=os.path.dirname(path))



import torch
from torch import nn
from torch.utils.data import DataLoader
from torch.optim.lr_scheduler import LinearLR, ReduceLROnPlateau
from potatorch.datasets.utils import split_dataset
from potatorch.datasets.utils import RandomSubsetSampler
from potatorch.datasets.utils import UnbatchedDataloader
from contextlib import ExitStack
import numpy as np
import gc

def make_optimizer(optimizer_init: callable, *args, **kwargs):
    return lambda m: optimizer_init(m.parameters(), *args, **kwargs)

class TrainingLoop():
    def __init__(self,
                 dataset,
                 loss_fn,
                 optimizer_fn,
                 train_p=0.7,
                 val_p=0.15,
                 test_p=0.15,
                 random_split=True,
                 batch_size=1024,
                 shuffle=False,
                 random_subsampling=None,
                 filter_fn=None,
                 num_workers=4,
                 mixed_precision=False,
                 callbacks=[],
                 val_metrics={},
                 device='cpu',
                 seed=42):
        self.dataset = dataset
        self.loss_fn = loss_fn
        self.optimizer_fn = optimizer_fn
        self.train_p = train_p
        self.val_p = val_p
        self.test_p = test_p
        self.random_split = random_split
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.random_subsampling = random_subsampling
        self.filter_fn = filter_fn
        self.num_workers = num_workers
        self.device = device
        self.mixed_precision = mixed_precision
        self.callbacks = callbacks
        self.val_metrics = val_metrics
        self.device = device
        self.seed = 42

        self._clear_state()
        self._init_dataloaders()

    def _clear_state(self):
        """ Clear internal training state """
        self.state = {}
        self.metrics = {}

    def _init_dataloaders(self):
        # TODO: maybe use an other class for evaluation?
        self.train_dataloader, self.val_dataloader, self.test_dataloader = self._make_dataloaders(
                self.dataset,
                self.train_p,
                self.val_p,
                self.test_p)

    def _collate_fn(self, batch):
        if self.filter_fn:
            batch = [data for data in batch if self.filter_fn(data)]
        return torch.utils.data.default_collate(batch)

    def _make_dataloaders(self, dataset, train_p, val_p, test_p):
        shuffle = self.shuffle and self.random_subsampling is not None

        train_ds, val_ds, test_ds = split_dataset(dataset,
                train_p, val_p, test_p, random=self.random_split,
                seed=self.seed)

        sampler = None
        if self.random_subsampling is not None:
            sampler = RandomSubsetSampler(train_ds, self.random_subsampling, replace=False)
        
        dataloader_fn = DataLoader if self.batch_size else UnbatchedDataloader

        train_dl = dataloader_fn(train_ds,
                batch_size=self.batch_size,
                shuffle=shuffle,
                sampler=sampler,
                collate_fn=self._collate_fn,
                pin_memory=True,
                num_workers=self.num_workers,
                prefetch_factor=(self.num_workers*2 if self.num_workers > 0 else None),
                # TODO: how to pass worker_init_fn
                # worker_init_fn=dataset.worker_init_fn,
                persistent_workers=False)
        val_dl = dataloader_fn(val_ds,
                batch_size=self.batch_size,
                collate_fn=self._collate_fn,
                shuffle=False,
                pin_memory=True,
                num_workers=self.num_workers,
                prefetch_factor=(self.num_workers*2 if self.num_workers > 0 else None),
                persistent_workers=self.num_workers > 0)
        test_dl = dataloader_fn(test_ds,
                batch_size=self.batch_size,
                collate_fn=self._collate_fn,
                shuffle=False,
                pin_memory=True,
                num_workers=0)

        return train_dl, val_dl, test_dl

    def _train(self, epochs):
        if self.train_dataloader is None or self.val_dataloader is None:
            self._init_dataloaders()

        num_batches = len(self.train_dataloader)
        self.update_state('batches', num_batches)
        
        self.on_train_start()

        # Pytorch auto scaler for mixed precision training
        if self.mixed_precision:
            scaler = torch.cuda.amp.GradScaler()

        initial_epoch = self.get_state('epoch', 1)

        for epoch in range(initial_epoch, initial_epoch + epochs):
            self.on_train_epoch_start(epoch)

            self.model.train()
            # TODO: provide interface to more than one input
            for batch, (X, y) in enumerate(self.train_dataloader):
                self.on_train_batch_start(batch)
                # TODO: generalize variable type
                # TODO: memory_format (e.g. torch.channels_last for 4D tensors)
                X = X.float().to(self.device, non_blocking=True)
                y = y.float().to(self.device, non_blocking=True)

                # Clear gradients
                self.optimizer.zero_grad(set_to_none=True)

                # Forward pass
                with ExitStack() as stack:
                    if self.mixed_precision:
                        stack.enter_context(torch.cuda.amp.autocast())
                pred = self.model(X).squeeze()
                loss = self.loss_fn(pred, y)

                # Backpropagation
                if self.mixed_precision:
                    scaler.scale(loss).backward()
                    scaler.step(self.optimizer)
                    scaler.update()
                else:
                    loss.backward()
                    self.optimizer.step()

                self.on_train_batch_end(batch, loss.detach())

            val_loss, other_metrics = self._test(self.val_dataloader, self.val_metrics)
            self.on_validation_end(val_loss, other_metrics)
            self.on_train_epoch_end(epoch)

        self.on_train_end()

    def _test(self, dataloader, metrics):
        num_batches = len(dataloader)
        self.model.eval()
        test_loss = 0.0
        test_metrics = dict.fromkeys(metrics.keys(), 0.0)
        # test_metrics = [0.0 for _ in metrics.keys()]
        with torch.no_grad():
            for X, y in dataloader:
                # TODO: generalize variable type
                X = X.float().to(self.device, non_blocking=True)
                y = y.float().to(self.device, non_blocking=True)

                pred = self.model(X).squeeze().detach()
                test_loss += self.loss_fn(pred, y).detach()
                for name, fn in metrics.items():
                    test_metrics[name] += fn(pred, y).detach()

        test_loss /= num_batches
        test_metrics = {k: v / num_batches for k, v in test_metrics.items()}
        return test_loss, test_metrics

    def predict(self, data):
        """ Run inference over a set of datapoints,
            returning the predicted values.
            `data` can be any iterable complying with the model input.
        """
        self.model.eval()
        h = []
        with torch.no_grad():
            for X in data:
                X = X.float().to(self.device, non_blocking=True)

                pred = self.model(X).squeeze()
                h = np.concatenate((h, pred.cpu().detach().numpy()), axis=None)

        return h
    
    def run(self, model, epochs=10, verbose=1):
        self.model = model.to(self.device)
        self.optimizer = self.optimizer_fn(self.model)
        self.update_state('verbose', verbose)
        try:
            self._train(epochs)
        except KeyboardInterrupt:
            print('Terminating training loop...')

        return self.model

    def clear(self):
        del self.train_dataloader
        del self.val_dataloader
        del self.test_dataloader
        self._clear_state()
    
    def dump_state(self):
        """ Produces a snapshot dictionary containing all information
            to restore the current state of the TrainingLoop sometime in the
            future.
        """
        assert self.model is not None, "Model not initialized"
        return {
                'model_state_dict': self.model.state_dict(),
                'optimizer_state_dict': self.optimizer.state_dict(),
                'training_loop_state_dict': self.get_states(),
                'training_loop_metrics': self.get_last_metrics(),
                'callbacks_state_dict': [c.state_dict() for c in self.callbacks],
                # 'train_dataloader': self.train_dataloader,
                # 'val_dataloader': self.val_dataloader,
                # 'test_dataloader': self.test_dataloader
                }

    def load_state(self, model, dump):
        """ Loads a TrainingLoop snapshot produced by a call to dump_state. """
        self.model = model
        self.model.load_state_dict(dump['model_state_dict'])
        self.optimizer.load_state_dict(dump['optimizer_state_dict'])
        self.state = dump['training_loop_state_dict']
        self.metrics = dump['training_loop_metrics']

        for i, d in enumerate(dump['callbacks_state_dict']):
            self.callbacks[i].load_state_dict(d)

    def get_last_metric(self, metric, default=None):
        """ Get last computed metric """
        return self.metrics.get(metric, default)

    def get_last_metrics(self):
        return self.metrics.copy()

    def update_metric(self, metric, value):
        """ Update the given metric with the current value
            The method automatically tracks min and max values of the metric
        """
        if torch.is_tensor(value):
            value = value.item()
        min_v = self.get_last_metric(f'min-{metric}')
        max_v = self.get_last_metric(f'max-{metric}')
        self.metrics[metric] = value
        if min_v is None or value < min_v:
            self.metrics[f'min-{metric}'] = value
        if max_v is None or value > max_v:
            self.metrics[f'max-{metric}'] = value

    def update_state(self, key, value):
        self.state[key] = value

    def get_state(self, key, default=None):
        return self.state.get(key, default)

    def get_states(self):
        return self.state.copy()

    """ Callback hooks """
    def on_train_start(self):
        for c in self.callbacks: c.on_train_start(self)

    def on_train_end(self):
        for c in self.callbacks: c.on_train_end(self)

    def on_train_batch_start(self, batch_num):
        self.update_state('batch', batch_num)
        for c in self.callbacks: c.on_train_batch_start(self)

    def on_train_batch_end(self, batch_num, batch_loss):
        # Update current batch loss
        self.update_metric('loss', batch_loss)
        # Current mean loss (default 0 if None)
        mean_loss = self.get_last_metric('mean_loss', 0.0)
        # Running mean loss update
        mean_loss = mean_loss + (batch_loss - mean_loss)/(batch_num + 1)
        self.update_metric('mean_loss', mean_loss)

        # Log current learning rate
        self.update_state('lr', self.optimizer.param_groups[0]['lr'])

        for c in self.callbacks: c.on_train_batch_end(self)

    def on_train_epoch_start(self, epoch_num):
        self.update_metric('mean_loss', 0.0)
        self.update_state('epoch', epoch_num)
        for c in self.callbacks: c.on_train_epoch_start(self)

    def on_train_epoch_end(self, epoch_num):
        # setting epoch + 1 on epoch end is necessary for cases in which
        # the training stopped mid-epoch
        self.update_state('epoch', epoch_num + 1)
        for c in self.callbacks: c.on_train_epoch_end(self)

    def on_validation_batch_start(self):
        for c in self.callbacks: c.on_validation_batch_start(self)

    def on_validation_batch_end(self):
        for c in self.callbacks: c.on_validation_batch_end(self)

    def on_validation_start(self):
        for c in self.callbacks: c.on_validation_start(self)

    def on_validation_end(self, val_loss, other_metrics):
        self.update_metric('val_loss', val_loss)

        for metric, value in other_metrics.items():
            self.update_metric(f'val_{metric}', value)
        for c in self.callbacks: c.on_validation_end(self)

    # TODO: maybe use an other class for evaluation?
    def evaluate(self):
        loss, other_metrics = self._test(self.test_dataloader, self.val_metrics)
        return {'loss': loss, **other_metrics}



import numpy as np
from scipy.stats import norm

class ValueEncoder():
  """ Encode categorical python objects into numbers """
  def __init__(self):
    self.map = {}
    self.code = 0
  
  def encode(self, value):
    """ Encode boolean and string values into unique integers """
    if isinstance(value, bool):
      return int(value)

    if not isinstance(value, str):
      return value
    
    if value not in self.map: 
      self.map[value] = self.code
      self.code += 1
    
    return self.map[value]

  def encode_dict(self, d):
    """ Encode every entry of a dictionary """
    new_d = {}
    for k in d.keys():
      new_d[k] = self.encode(d[k])
    return new_d

class GaussianProcessRegressor():
  def __init__(self, sigma):
    self.sigma = sigma
    self.sigma2 = sigma**2
  
  def k(self, x1, x2):
    """ Gaussian kernel """
    return np.exp(-10/2 * np.linalg.norm(x1 - x2)**2)

  def gram(self, xs1, xs2):
    """ Gram (covariance) matrix """
    return np.array([[self.k(x1, x2) for x2 in xs2] for x1 in xs1])
    
  def fit(self, x, y):
    """ Fit the model with the given dataset D = {(x,y)}"""
    self.x_train = x
    self.y_train = y
    self.C = self.gram(x, x) + self.sigma2*np.eye(len(x))
    self.C_inv = np.linalg.inv(self.C)
  
  def predict(self, x):
    """ Perform inference on a new set of data points """
    K = self.gram(self.x_train, x)
    mu = K.T.dot(self.C_inv).dot(self.y_train)
    mu = mu.reshape((len(x),))
    cov = self.gram(x, x) - K.T.dot(self.C_inv).dot(K)
    return mu, cov

  def sample(self, x, n=1):
    """ Sample a function from the posterior distribution modeled by the GP """
    mu, cov = self.predict(x)
    return np.random.multivariate_normal(mu, cov, size=n)

class BayesianOptimizer():
  """ Bayesian function optimizer:
    Perform optimization in a bayesian setting by modelling the function's
    likelihood distribution with a Gaussian Process, which is then used as a
    surrogate function to optimize the original function.
    The acquisition function determines the next point to evaluate.
    New evaluated points can be provided to the optimizer with the `update`
    method, which will in turn improve the Bayesian model's beliefs.
    -----------------------------------
    Parameters:
  - initial_x:   intial configuration of hyperparameters (dictionary object)
  - initial_y:   value of score_function(initial_x) (scalar value)
  - minimize:    whether to minimize or maximize the function
  """
  def __init__(self, initial_x, initial_y, minimize=True, sigma=1e-1):
    self.gp = GaussianProcessRegressor(sigma=sigma)
    self.encoder = ValueEncoder()
    # Convert initial_x into a vector of numbers (see ValueEncoder)
    self.x = np.array([list(self.encoder.encode_dict(initial_x).values())])
    self.y = np.array([initial_y])
    self.minimize = minimize
    self.gp.fit(self.x, self.y)

  def acquisition(self, x_sampled):
    """ Acquisition function: excepted improvement """
    mu, _ = self.gp.predict(self.x)
    best = np.min(mu) if self.minimize else np.max(mu)
    mu, cov = self.gp.predict(x_sampled)
    mu = mu.reshape((len(x_sampled),))
    std = np.sqrt(np.diag(cov))
    sign = 1 if self.minimize else -1
    Z = lambda mu, std: sign*(best - mu)/(std**2) if std > 0 else 0.0

    improvements = [
             sign*(best - mu_x)*norm.cdf(Z(mu_x, std_x)) +
             (std_x**2)*norm.pdf(Z(mu_x, std_x))
              for mu_x, std_x in zip(mu, std)
               ]
    return improvements

  def optimize(self, sample):
    """ Perform an iteration of bayesian optimization """
    # Convert each sample into a vector of numbers
    encoded = np.array([list(self.encoder.encode_dict(x).values()) for x in sample])
    scores = self.acquisition(encoded)
    best_ix = np.argmax(scores) # always take the max (maximizing expectation of improvement)
    query = sample[best_ix]
    return query
  
  def update(self, x, y):
    """ Update the bayesian model's belief """
    x = np.array(list(self.encoder.encode_dict(x).values()))
    self.x = np.vstack((self.x, [x]))
    self.y = np.vstack((self.y, [y]))
    self.gp.fit(self.x, self.y)



import torch
from torch import nn

class MSRELoss(nn.Module):
    """ Implements Mean Squared Relative Error Loss """
    def __init__(self, delta=1):
        super(MSRELoss, self).__init__()
        self.delta = delta

    def forward(self, input, target):
        # return torch.mean(torch.div(torch.abs(input - target), torch.abs(target) + self.delta))
        return torch.mean((torch.log(torch.clamp(input, min=(-self.delta + 1)) + self.delta) - torch.log(torch.clamp(target, min=(-self.delta + 1)) + self.delta))**2)

        # return torch.mean(torch.div((input - target)**4, target**2 + self.eps))
        # return torch.mean(torch.div((input - target)**2, target**2 + 1e-2))
        # return torch.mean(torch.log((input - target)**2))
        # return torch.mean((input - target)**2)
        # return torch.mean(torch.log(torch.cosh(input - target)))



import torch
from torch import nn

from potatorch.training import TrainingLoop, make_optimizer
from potatorch.callbacks import ProgressbarCallback
from torch.utils.data import TensorDataset

# Fix a seed for TrainingLoop to make non-deterministic operations such as
# shuffling reproducible
SEED = 42
device = 'cuda'

epochs = 100
lr = 1e-4

# Define your model as a pytorch Module
model = nn.Sequential(nn.Linear(1, 128), nn.ReLU(), 
        nn.Linear(128, 128), nn.ReLU(),
        nn.Linear(128, 1))

# Create your dataset as a torch.data.Dataset
dataset = TensorDataset(torch.arange(1000).view(1000, 1), torch.sin(torch.arange(1000)))

# Provide a loss function and an optimizer
loss_fn = torch.nn.MSELoss()
optimizer = make_optimizer(torch.optim.Adam, lr=lr)

# Construct a TrainingLoop object.
# TrainingLoop handles the initialization of dataloaders, dataset splitting,
# shuffling, mixed precision training, etc.
# You can provide callback handles through the `callbacks` argument.
training_loop = TrainingLoop(
        dataset,
        loss_fn,
        optimizer,
        train_p=0.8,
        val_p=0.1,
        test_p=0.1,
        random_split=False,
        batch_size=None,
        shuffle=False,
        device=device,
        num_workers=0,
        seed=SEED,
        val_metrics={'l1': nn.L1Loss(), 'mse': nn.MSELoss()},
        callbacks=[
            ProgressbarCallback(epochs=epochs, width=20),
            ]
        )
# Run the training loop
model = training_loop.run(model, epochs=epochs)



import torch
from torch.utils.data import random_split, Sampler, Dataset
from torch.utils.data.dataset import Subset
from typing import Sized, Iterator
import numpy as np
import random
from torch import default_generator, randperm
from torch._utils import _accumulate

def memory_safe_random_split(dataset, lengths, generator=default_generator, random=True):
    """ same as torch.utils.data.random_split, but passes a `ndarray` to Subset
    in order to avoid memory leaks caused by python refcounting behaviour of
    native lists
    """
    # Cannot verify that dataset is Sized
    if sum(lengths) != len(dataset):    # type: ignore[arg-type]
        raise ValueError("Sum of input lengths does not equal the length of the input dataset!")

    if random:
        indices = np.array(randperm(sum(lengths), generator=generator).tolist())
    else:
        indices = np.arange(sum(lengths))
        
    return [Subset(dataset, indices[offset - length : offset]) for offset, length in zip(_accumulate(lengths), lengths)]

def split_dataset(ds, train_p=0.7, val_p=0.15, test_p=0.15, random=True, seed=42):
    """ Split a torch.utils.data.Datset into 3 random split, according to the
        given percentages.

        train_p, val_p, test_p must sum up to 1

        returns (train_ds, val_ds, test_ds)
    """
    train_size = int(len(ds) * train_p)

    # if ds.oversample:
    #     val_size = int(len(ds) * val_p)
    #     test_size = len(ds) - train_size - val_size
    # else:
    val_size = int(len(ds) * val_p)
    test_size = len(ds) - train_size - val_size

    train_ds, val_ds, test_ds = memory_safe_random_split(ds,
            [train_size, val_size, test_size],
            generator=torch.Generator().manual_seed(seed),
            random=random
            )

    # for clarity
    return train_ds, val_ds, test_ds

class RandomSubsetSampler(Sampler[int]):
    def __init__(self, data_source: Sized, fraction: int, replace=False) -> None:
        self.fraction = fraction
        self.data_source = data_source
        self.replace = replace

    def __iter__(self) -> Iterator[int]:
        n = len(self.data_source)
        if not self.replace:
            yield from iter(np.random.choice(n, int(n*self.fraction), replace=False))
        else:
            for i in range(int(n * self.fraction)):
                yield np.random.randint(low=0, high=n)

    def __len__(self) -> int:
        return int(len(self.data_source) * self.fraction)

class UnbatchedDataloader():
    def __init__(self, dataset, *args, **kwargs):
        self.dataset = dataset

    def __iter__(self):
        self.consumed = False
        return self

    def __next__(self):
        if self.consumed:
            raise StopIteration
        self.consumed = True
        return self.dataset[:]

    def __len__(self):
        return min(1, len(self.dataset))



import torch
from torch import nn

from potatorch.training import TrainingLoop, make_optimizer
from potatorch.callbacks import ProgressbarCallback
from potatorch.optimization.tuning import HyperOptimizer
from torch.utils.data import TensorDataset
from functools import partial

def train(dataset, device, config):
    """ Your usual training function that runs a TrainingLoop instance """
    # Fix a seed for TrainingLoop to make non-deterministic operations such as
    # shuffling reproducible
    SEED = 42
    # NOTE: `epochs` is a fixed hyperparameter; it won't change among runs
    epochs = config['epochs']

    # Define your model as a pytorch Module
    model = nn.Sequential(nn.Linear(1, 128), nn.ReLU(), 
            nn.Linear(128, 128), nn.ReLU(),
            nn.Linear(128, 1))

    loss_fn = torch.nn.MSELoss()
    # Provide a loss function and an optimizer
    # NOTE: `lr` is a dynamic hyperparameter; it will change among runs
    optimizer = make_optimizer(torch.optim.Adam, lr=config['lr'])

    # Construct a TrainingLoop object.
    # TrainingLoop handles the initialization of dataloaders, dataset splitting,
    # shuffling, mixed precision training, etc.
    # You can provide callback handles through the `callbacks` argument.
    training_loop = TrainingLoop(
            dataset,
            loss_fn,
            optimizer,
            train_p=0.8,
            val_p=0.1,
            test_p=0.1,
            random_split=False,
            batch_size=None,
            shuffle=False,
            device=device,
            num_workers=0,
            seed=SEED,
            val_metrics={'l1': nn.L1Loss(), 'mse': nn.MSELoss()},
            callbacks=[
                ProgressbarCallback(epochs=epochs, width=20),
                ]
            )
    # Run the training loop
    model = training_loop.run(model, epochs=epochs, verbose=1)
    # Return a dictionary containing the training and validation metrics 
    # calculated during the last epoch of the loop
    return training_loop.get_last_metrics()

# Define your search configuration
search_config = {
        'method': 'grid',   # which search method to use: ['grid', 'bayes', 'random']
        'metric': {
            'name': 'val_loss', # the metric you're optimizing
            'goal': 'minimize'  # whether you want to minimize or maximize it
        },
        'parameters': { # the set of hyperparameters you want to optimize
            'lr': {
                'values': [1e-2, 1e-3, 1e-4]    # a range of values for the grid search to try
            }
        },
        'fixed': {      # fixed hyperparameters that won't change among runs
            'epochs': 200
        }
    }

def main():
    device = 'cuda'
    # Create your dataset as a torch.data.Dataset
    dataset = TensorDataset(torch.arange(1000).view(1000, 1), torch.sin(torch.arange(1000)))
    # Apply additional parameters to the train function to have f(config) -> {}
    score_function = partial(train, dataset, device)
    # Construct the hyperparameters optimizer
    hyperoptimizer = HyperOptimizer(search_config)
    # Run the optimization over the hyperparameters space
    config, error = hyperoptimizer.optimize(score_function, return_error=True)
    print('Best configuration found: {}\n with error: {}'.format(config, error))

if __name__ == '__main__':
    main()



import os
import pkbar # progress bar for pytorch
import wandb
import torch
import numpy as np
from torch import nn
from torch.optim.lr_scheduler import LinearLR, ReduceLROnPlateau
from torch.optim.lr_scheduler import CosineAnnealingWarmRestarts
from torch.optim.lr_scheduler import CosineAnnealingLR
from abc import ABC, abstractmethod
from potatorch.training import TrainingLoop

class TrainingCallback(ABC):
    """ Training Callback base class """
    def __init__(self):
        return
    
    @abstractmethod
    def state_dict(self):
        pass

    @abstractmethod
    def load_state_dict(self, state_dict):
        pass

    def on_train_start(self, state):
        return

    def on_train_end(self, state):
        return

    def on_train_batch_start(self, state):
        return

    def on_train_batch_end(self, state):
        return

    def on_train_epoch_start(self, state):
        return

    def on_train_epoch_end(self, state):
        return

    def on_validation_batch_start(self, state):
        return

    def on_validation_batch_end(self, state):
        return

    def on_validation_start(self, state):
        return

    def on_validation_end(self, state):
        return

class ProgressbarCallback(TrainingCallback):
    kbar: pkbar.Kbar

    def __init__(self, epochs, width=20):
        super().__init__()
        self.epochs = epochs
        self.width = width

    def state_dict(self):
        return None

    def load_state_dict(self, state_dict):
        pass

    def on_train_epoch_start(self, state):
        super().on_train_epoch_start(state)
        # reset incremental averages
        self.loss = 0.0
        self.lr = 0.0

        # pkbar expects epochs starting from zero
        epoch = state.get_state('epoch', 1) - 1
        num_batches = state.get_state('batches')
        verbose = state.get_state('verbose')
        ################################### Initialization ########################################
        if verbose and epoch is not None and num_batches is not None:
            self.kbar = pkbar.Kbar(
                    target=num_batches,
                    num_epochs=self.epochs,
                    epoch=epoch,
                    width=self.width,
                    always_stateful=True)
        # By default, all metrics are averaged over time. If you don't want this behavior, you could either:
        # 1. Set always_stateful to True, or
        # 2. Set stateful_metrics=["loss", "rmse", "val_loss", "val_rmse"], Metrics in this list will be displayed as-is.
        # All others will be averaged by the progbar before display.
        ###########################################################################################
    
    def on_train_batch_end(self, state):
        super().on_train_batch_end(state)
        # TODO: use self.metrics and self.state to update the progress bar
        batch = state.get_state('batch')
        verbose = state.get_state('verbose')
        self.lr = self.lr + (state.get_state('lr') - self.lr) / (batch + 1)
        self.loss = self.loss + (state.get_last_metric('loss') - self.loss) / (batch + 1)
        if verbose and batch is not None and self.loss and self.lr:
            self.kbar.update(batch, values=[('loss', self.loss), ('lr', self.lr)])

    def on_validation_end(self, state):
        super().on_validation_end(state)
        # TODO: use self.metrics to update the progress bar
        val_loss = state.get_last_metric('val_loss')
        verbose = state.get_state('verbose')
        other_metrics = state.val_metrics.keys()
        if verbose and val_loss is not None:
            # formatting rule for validation metrics: val_{metric}
            val_m = lambda m: f'val_{m}'
            self.kbar.add(1, values=[('val_loss', val_loss), 
                # retrieve and zip other validation metrics from the state
                *[(val_m(m), state.get_last_metric(val_m(m))) for m in other_metrics]
                ])

class LRSchedulerCallback(TrainingCallback):
    def __init__(self, optimizer, warmup_steps=1000, cosine_annealing=True, restart=False, cosine_tmax=None, cosine_factor=None, min_lr=0.0):
        super().__init__()
        self.optimizer = optimizer
        self.warmup_steps = warmup_steps
        self.lr_warmup = LinearLR(self.optimizer, start_factor=0.001, total_iters=self.warmup_steps)
        self.lr_decay = ReduceLROnPlateau(self.optimizer, mode='min', factor=0.5, patience=5)
        self.lr_cosine = None
        self.cosine_annealing = cosine_annealing
        self.cosine_tmax = cosine_tmax
        self.cosine_factor = cosine_factor
        self.restart = restart
        self.min_lr = min_lr

        if self.cosine_tmax is None and self.cosine_annealing:
            self.cosine_tmax = 50
            if not self.cosine_factor:
                self.cosine_factor = 1
            self.cosine_factor = int(self.cosine_factor)

    def _init_cosine_annealing(self):
        # With restarts
        if self.restart:
            self.lr_cosine = CosineAnnealingWarmRestarts(self.optimizer, self.cosine_tmax, self.cosine_factor, eta_min=self.min_lr)
        # Without restarts
        else:
            self.lr_cosine = CosineAnnealingLR(self.optimizer, self.cosine_tmax, eta_min=self.min_lr)

    def state_dict(self):
        state_dict = {}
        state_dict.update({
            'lr_warmup_state_dict': self.lr_warmup.state_dict(),
            'lr_decay_state_dict': self.lr_decay.state_dict()
            })

        if self.lr_cosine:
            state_dict.update({
                'lr_cosine_state_dict': self.lr_cosine.state_dict()
                })

        return state_dict

    def load_state_dict(self, state_dict):
        self.lr_warmup.load_state_dict(state_dict['lr_warmup_state_dict'])
        self.lr_decay.load_state_dict(state_dict['lr_decay_state_dict'])

        if self.cosine_annealing:
            if self.lr_cosine is None:
                self._init_cosine_annealing()
            self.lr_cosine.load_state_dict(state_dict.get('lr_cosine_state_dict', {}))
    
    def on_train_start(self, state):
        super().on_train_start(state)
        state.update_state('lr', self.lr_warmup.get_last_lr()[0])

    def on_train_batch_end(self, state):
        super().on_train_batch_end(state)
        self.lr_warmup.step()
        state.update_state('lr', self.lr_warmup.get_last_lr()[0])

    def on_validation_end(self, state):
        super().on_validation_end(state)
        val_loss = state.get_last_metric('val_loss')

        # Cosine annealing
        if self.cosine_annealing:
            batches = state.get_state('batches')
            epoch = state.get_state('epoch')
            # Only when warmup is over
            if epoch is not None and batches and epoch * batches > self.warmup_steps:
                # Init cosine annealing if None
                if self.lr_cosine is None:
                    self._init_cosine_annealing()

                # Apply cosine annealing
                self.lr_cosine.step()
        # Decay on plateau (if cosine_annealing is False)
        elif val_loss is not None:
            self.lr_decay.step(val_loss)
        state.update_state('lr', self.optimizer.param_groups[0]['lr'])


class WandbCallback(TrainingCallback):
    def __init__(self, project_name, entity, config=None, tags=None, save_code=True, log=True, batch_frequency=None):
        super().__init__()
        self.project_name = project_name
        self.entity = entity
        self.tags = tags
        self.config = config
        self.save_code = save_code
        self.log = log
        self.batch_frequency = batch_frequency
        self.run_id = None

    def state_dict(self):
        state_dict = {}
        state_dict.update({'run_id': self.run_id})
        return state_dict

    def load_state_dict(self, state_dict):
        self.run_id = state_dict.get('run_id', None)

    def init(self):
        return wandb.init(
                id=self.run_id,
                project=self.project_name,
                entity=self.entity,
                config=self.config,
                resume='must' if self.run_id else None,
                tags=self.tags,
                save_code=self.save_code,
                reinit=True)


    def on_train_start(self, state):
        super().on_train_start(state)
        # Create a new wandb run
        self.run = self.init()
        self.run_id = self.run.id

    def on_train_end(self, state):
        super().on_train_end(state)
        # Stop the current run
        if self.run:
            self.run.finish()

    def on_train_epoch_end(self, state):
        super().on_train_epoch_end(state)
        # We can probably do everything after validation
        # if self.log:
        #     wandb.log(state.get_states() | state.get_last_metrics(), commit=True)

    def on_train_batch_end(self, state):
        super().on_train_batch_end(state)
        batch = state.get_state('batch')
        if self.log and self.batch_frequency and batch \
                    and batch % self.batch_frequency == 0:
            wandb.log({**state.get_states(), **state.get_last_metrics()}, commit=True)

    def on_validation_end(self, state):
        super().on_validation_end(state)
        if self.log:
            wandb.log({**state.get_states(), **state.get_last_metrics()}, commit=True)


class CheckpointCallback(TrainingCallback):
    def __init__(self, path, save_best=True, detect_anomaly=False, metric='val_loss', frequency=None, mode='min', sync_wandb=False, debug=False):
        super().__init__()
        self.path = path
        self.save_best = save_best
        self.detect_anomaly = detect_anomaly
        self.metric = metric
        self.frequency = frequency
        self.mode = mode
        self.sync_wandb = sync_wandb
        self.debug = debug
        self.anomaly_detected = False

        assert path, 'path must not be None'
        assert save_best or frequency or detect_anomaly, 'Either one among save_best, frequency and detect_anomaly must be provided'
        assert not (save_best and frequency), 'Only one among save_best, frequency and detect_anomaly can be provided'
        assert not (save_best and detect_anomaly), 'Only one among save_best, frequency and detect_anomaly can be provided'
        assert not (frequency and detect_anomaly), 'Only one among save_best, frequency and detect_anomaly can be provided'
        assert (not save_best) or (mode == 'min' or mode == 'max'), "If save_best is True, then mode can either be 'min' or 'max'"
        assert (not save_best) or metric, 'If save_best = True, then a metric must be provided'

        if self.save_best:
            self.best = np.inf if mode == 'min' else -np.inf
            self.default = self.best

        if self.save_best and self.mode:
            self.minimize = mode == 'min'

    def state_dict(self):
        state_dict = {}
        if self.save_best:
            state_dict.update({'best': self.best})
        return state_dict

    def load_state_dict(self, state_dict):
        if 'best' in state_dict:
            self.best = state_dict['best']

    def on_train_start(self, state):
        super().on_train_start(state)

    def on_train_end(self, state):
        super().on_train_end(state)

    def on_train_epoch_end(self, state):
        super().on_train_epoch_end(state)
        epoch = state.get_state('epoch', 1)
        output = False

        # Save model checkpoint after each epoch, until an anomaly is detected
        if self.detect_anomaly and not self.anomaly_detected:
            metric = state.get_last_metric(self.metric)
            if torch.isnan(metric) or torch.isinf(metric):
                self.anomaly_detected = True
                print('Anomaly detected, disabling model checkpointing')
            else:
                self._save_checkpoint(state)
                output = True

        if self.frequency and epoch % self.frequency == 0:
            self._save_checkpoint(state)
            output = True

        if self.save_best:
            current = state.get_last_metric(self.metric, self.default)
            if (self.minimize and current < self.best) or \
               (not self.minimize and current > self.best):
                   self.best = current
                   self._save_checkpoint(state)
                   output = True
        if self.debug and output:
            print('Saving model checkpoint')

    def _save_checkpoint(self, state):
        """ Save the current state checkpoint to the specified self.path """
        # Save dump to disk
        torch.save(state.dump_state(), self.path)
        # Sync file with wandb
        if self.sync_wandb:
            wandb.save(self.path, base_path=os.path.dirname(self.path))


class SanityCheckCallback(TrainingCallback):
    def __init__(self, data, descriptors=None, transform=None, target_transform=None):
        self.data = data
        self.descriptors = descriptors
        self.transform = transform
        self.target_transform = target_transform

        if self.transform:
            self.data = [self._rebatch(self.transform(x[0], aux[0])) for (x, aux) in self.data]

        self.data = self._make_tensors(self.data)

    def _rebatch(self, data):
        return (np.expand_dims(data[0], 0), np.expand_dims(data[1], 0))

    def _make_tensors(self, data):
        return [(torch.from_numpy(x), torch.from_numpy(aux)) for (x, aux) in data]

    def state_dict(self):
        return {}

    def load_state_dict(self, state_dict):
        pass

    def on_train_epoch_end(self, state):
        h = state.predict(self.data)
        if self.descriptors is not None:
            samples = zip(self.descriptors, self.data, h)
        else:
            samples = enumerate(zip(self.data, h))
        for (i, x, y) in samples:
            if self.target_transform:
                y = self.target_transform(y)
            print(f'{i}: {y}')
